import os
import pysam
import random
import tempfile
import multiprocessing
from functools import partial
from .utils import ensure_bam_index, chunk_bam, process_bam_chunk_nt_qual, process_bam_chunk_nt_count, parse_filename
import glob
                            
def run_bam_to_tsv(args):
  """
  Given a BAM file and ref fasta, convert it to a TSV file with nucleotide quality counts or nucleotide variant counts depending on args.type.
  Output a TSV file with the following columns (--type nt_qual):
    - chr
    - start (0-based)
    - end
    - ref_nt
    - qual_0_10: nt qualities falling into this range [)
    - qual_10_20 [)
    - qual_20_30 [)
    - qual_30_40 [)
    - qual_40_above [)
  Output a TSV file with the following columns (--type nt_count):
    - chr
    - start (0-based)
    - end
    - ref_nt
    - A (number of 'A' nt)
    - T (number of 'T' nt)
    - C (number of 'C' nt)
    - G (number of 'G' nt)
    - N (number of all other nt)
  """
  
  # Check input
  if not os.path.exists(args.bam):
    raise FileNotFoundError(f"The input BAM file {args.bam} does not exist.")
  ensure_bam_index(args.bam)
  if not os.path.exists(args.ref):
    raise FileNotFoundError(f"The reference FASTA file {args.ref} does not exist.")

  # Chunk bam for multiprocessing
  chunks = []
  with pysam.AlignmentFile(args.bam, "rb") as bamfile:
    ref_names = (bamfile.references)
    ref_lengths = [bamfile.get_reference_length(x) for x in ref_names]

    for ref_name, ref_length in zip(ref_names, ref_lengths):
      chunks.extend(chunk_bam(ref_name, ref_length, args.ncpus))
    random.shuffle(chunks) # ensure each parallel job load is even (some chr region may have more coverage)
    
  # Process each chunk in parallel
  with tempfile.TemporaryDirectory() as temp_dir:
    print('Temp folder path: ', temp_dir)
    with multiprocessing.Pool(processes = args.ncpus) as pool:
      if args.type == 'nt_qual':
        header = '\t'.join(['#chr', 'start', 'end', 'ref_nt', 'qual_0_10', 'qual_10_20', 'qual_20_30', 'qual_30_40', 'qual_40_above', 'qual_avg']) + '\n'
        func = partial(process_bam_chunk_nt_qual, bam_path = args.bam, ref_path = args.ref, temp_dir = temp_dir)
      elif args.type == 'nt_count':
        header = '\t'.join(['#chr', 'start', 'end', 'ref_nt', 'A', 'T', 'C', 'G', 'N']) + '\n'
        func = partial(process_bam_chunk_nt_count, bam_path = args.bam, ref_path = args.ref, temp_dir = temp_dir)
      pool.starmap(func, chunks)
      
      tsv_files = glob.glob(os.path.join(temp_dir, '*.tsv'))
      tsv_sorted = sorted(tsv_files, key = lambda f: (parse_filename(f)[0], parse_filename(f)[1]))
      if not os.path.exists(os.path.dirname(args.output)):
        os.makedirs(os.path.dirname(args.output))
      with pysam.BGZFile(args.output, 'w') as out:
        out.write(header.encode())

        for path in tsv_sorted:
          with open(path) as f:
            for line in f:
                if line.strip():
                  out.write(line.encode())


import argparse
from .utils import validate_range

def str2bool(v):
  if isinstance(v, bool):
    return v
  if v.lower() in ('yes', 'true', 't', 'y', '1'):
    return True
  elif v.lower() in ('no', 'false', 'f', 'n', '0'):
    return False
  else:
    raise argparse.ArgumentTypeError('Boolean value expected.')

def create_args():
  parser = argparse.ArgumentParser(description = 'nanoloop')
  subparsers = parser.add_subparsers(dest = 'command')
  
  # Add version command
  try:
    from . import __version__
    version = __version__
  except ImportError:
    version = "unknown"
  parser.add_argument('--version',
                     action = 'version',
                     version = version)

  # subcommand: bam_to_tsv
  parser_bam_to_tsv = subparsers.add_parser('bam_to_tsv',
                                            help = 'Parse BAM file and output TSV file')
  parser_bam_to_tsv.add_argument('--bam', 
                                  type = str, 
                                  required = True,
                                  help = 'Path to input BAM file')
  parser_bam_to_tsv.add_argument('--ref', 
                                  type = str, 
                                  required = True,
                                  help = 'Path to reference fasta file')
  parser_bam_to_tsv.add_argument('--output', 
                                  type = str, 
                                  required = True,
                                  help='Path to output tsv.gz file')
  parser_bam_to_tsv.add_argument('--type', 
                                  type = str, 
                                  required = True,
                                  choices = ['nt_count', 'nt_qual'],
                                  help = 'Type of tsv file to output: "nt_count" or "nt_qual", the former is for nucleotide species count at each site and the latter is for nucleotide read quality count at each site')
  parser_bam_to_tsv.add_argument('--ncpus', 
                                  type = int, 
                                  default = 1,
                                  help = 'Number of CPUs to use')
  
  # subcommand: tsv_to_plot
  parser_tsv_to_plot = subparsers.add_parser('tsv_to_plot',
                                            help = 'Parse TSV file and output plot')
  parser_tsv_to_plot.add_argument('--tsv', 
                                  type = str, 
                                  required = True,
                                  help = 'Path to input TSV file')
  parser_tsv_to_plot.add_argument('--type', 
                                  type = str, 
                                  required = True,
                                  choices = ['nt_count', 'nt_qual'],
                                  help = 'Type of input TSV file being used: "nt_count" or "nt_qual", the former is for nucleotide species count at each site and the latter is for nucleotide read quality count at each site')
  parser_tsv_to_plot.add_argument('--range', 
                                  type = str, 
                                  required = True,
                                  help = 'Range to plot: e.g. "chr1:1000000-1100000"')
  parser_tsv_to_plot.add_argument('--mode', 
                                  type = str, 
                                  choices = ['ratio', 'count'],
                                  default = 'ratio', 
                                  help = 'Plot mode: "ratio" or "count", the former is for C_to_T fraction and the latter is for raw count at each site')
  parser_tsv_to_plot.add_argument('--add_gc', 
                                  type = str2bool, 
                                  default = False,
                                  help = 'Add GC content to plot, note that GC content is calculated from the TSV file, which may not cover all reference positions')
  parser_tsv_to_plot.add_argument('--add_qual_avg', 
                                  type = str2bool, 
                                  default = False,
                                  help = 'Relevant when --type is "nt_qual", add rolling average of quality score to plot, mutual exclusive to --add_gc')
  parser_tsv_to_plot.add_argument('--output', 
                                  type = str, 
                                  required = True,
                                  help='Path to output pdf plot file, must be ended with ".pdf" or ".png"')
  
  # subcommand: tsv_to_bed
  parser_tsv_to_bed = subparsers.add_parser('tsv_to_bed',
                                            help = 'Convert TSV file to BED format for MACS3 peak calling. For "--type nt_qual": regions with lower quality scores will generate more simulated read tags, creating peaks in those regions. For "--type nt_count": the number of simulated read tags is proportional to the fraction of C converted to T at each reference C position')
  parser_tsv_to_bed.add_argument('--tsv', 
                                  type = str, 
                                  required = True,
                                  help = 'Path to input TSV file')
  parser_tsv_to_bed.add_argument('--output',  
                                  type = str, 
                                  required = True,
                                  help = 'Path to output bed.gz file')
  parser_tsv_to_bed.add_argument('--scale', 
                                  type = float, 
                                  default = 1.0,
                                  help = 'Scale factor for quality score, value larger than 1 will make the read coverage difference simulated from quality score larger, and the resulting bed will have more simulated read tags in it')
  parser_tsv_to_bed.add_argument('--type', 
                                  type = str, 
                                  required = True,
                                  choices = ['nt_count', 'nt_qual'],
                                  help = 'Type of input TSV file being used: "nt_count" or "nt_qual", the former is for nucleotide species count at each reference C site and the latter is for nucleotide read quality count at each reference site')
  
  # subcommand: tsv_to_peak
  parser_tsv_to_peak = subparsers.add_parser('tsv_to_peak',
                                            help = 'Call peaks using a sliding window approach')
  parser_tsv_to_peak.add_argument('--tsv', 
                                  type = str, 
                                  required = True,
                                  help = 'Path to input TSV file, both .tsv and .tsv.gz are supported, if .tsv.gz, the file size must be less than your memory size')
  parser_tsv_to_peak.add_argument('--type', 
                                  type = str, 
                                  required = True,
                                  choices = ['nt_qual', 'nt_count'],
                                  help = 'Type of input TSV file being used: "nt_qual" or "nt_count", the former is for nucleotide read quality count at each reference site and the latter is for nucleotide species count at each reference C site')
  parser_tsv_to_peak.add_argument('--low_qual_cutoff', 
                                  type = int, 
                                  default = 30,
                                  choices = [10, 20, 30, 40],
                                  help = 'Low quality cutoff, relevant when --type is "nt_qual", choose from 10, 20, 30, and 40, quality score below this cutoff is considered low quality, default to 30')
  parser_tsv_to_peak.add_argument('--frac_cutoff', 
                                  type = float, 
                                  default = 0.4,
                                  help = 'Fraction cutoff for peak calling, relevant when --type is "nt_qual"; given a window size, the fraction of low quality reads in the window is calculated, if the fraction is larger than the cutoff, the window is considered a peak, default to 0.4')
  parser_tsv_to_peak.add_argument('--conversion_cutoff', 
                                  type = float, 
                                  default = 0.05,
                                  help = 'Fraction cutoff for miscalled T at reference C site, relevant when --type is "nt_count", used to call peaks at reference C sites')
  parser_tsv_to_peak.add_argument('--min_peak_length', 
                                  type = int, 
                                  default = 0,
                                  help = 'Minimum length of peaks to keep, default to 50, detected peaks shorter than this value will be discarded')
  parser_tsv_to_peak.add_argument('--window_size', 
                                  type = int, 
                                  default = 50,
                                  help = 'Window size for peak calling, default to 50') 
  parser_tsv_to_peak.add_argument('--merge_nearby_peaks', 
                                  type = str2bool, 
                                  default = True,
                                  help = 'Merge nearby peaks, peaks that are within the (2 * window size) will be merged into one peak, default to True')
  parser_tsv_to_peak.add_argument('--output', 
                                  type = str, 
                                  required = True,
                                  help = 'Path to output bed.gz file')
  parser_tsv_to_peak.add_argument('--ncpus', 
                                  type = int, 
                                  default = 1,
                                  help = 'Number of CPUs to use, larger number may increase the memory usage significantly')
  
  args = parser.parse_args()
  
  if args.command == 'bam_to_tsv':
    if not args.output.endswith('.tsv.gz'):
      parser.error('The --output path must end with ".tsv.gz"!')
  
  if args.command == 'tsv_to_plot':
    if not args.tsv.endswith('.tsv.gz'):
      parser.error('The --tsv path must end with .tsv.gz and must be created with bgzip!')
    validate_range(args.range)
    if args.type == 'nt_qual':
      if args.add_gc and args.add_qual_avg:
        parser.error('Only one of --add_gc and --add_qual_avg can be specified at a time!')
    if args.type == 'nt_count' and args.add_qual_avg:
      parser.error('--add_qual_avg is not applicable for --type nt_count!')

  if args.command == 'tsv_to_bed':
    if not args.tsv.endswith('.tsv.gz'):
      parser.error('The --tsv path must end with .tsv.gz and must be created with bgzip!')
    if not args.output.endswith('.bed.gz'):
      parser.error('The --output path must end with .bed.gz!')
  
  if args.command == 'tsv_to_peak':
    if not args.output.endswith('.bed.gz'):
      parser.error('The --output path must end with .bed.gz!')

  return args


from .arg_parser import create_args
from .bam_to_tsv import run_bam_to_tsv
from .tsv_to_plot import run_tsv_to_plot
from .tsv_to_bed import run_tsv_to_bed
from .tsv_to_peak import run_tsv_to_peak

def main():
  args = create_args()

  if args.command == 'bam_to_tsv':
    run_bam_to_tsv(args)
  elif args.command == 'tsv_to_plot':
    run_tsv_to_plot(args)
  elif args.command == 'tsv_to_bed':
    run_tsv_to_bed(args)
  elif args.command == 'tsv_to_peak':
    run_tsv_to_peak(args)

if __name__ == '__main__':
  main()


import os
from .utils import ensure_tsv_index, extract_range_from_tsv
import seaborn as sns
import numpy as np  
import matplotlib.pyplot as plt
def run_tsv_to_plot(args):
  """
  Given tsv.gz from bam_to_tsv, parse out desired region and plot.
  """
  
  # Check input
  if not os.path.exists(args.tsv):
    raise FileNotFoundError(f"The input TSV file {args.tsv} does not exist.")
  ensure_tsv_index(args.tsv)

  # Extract desired region 
  df = extract_range_from_tsv(args.tsv, args.range)
  
  if args.type == 'nt_count':
    # Calculate ref_nt C converted to all other nucleotides ratio
    if (args.mode == "ratio"):
      y_label = 'Allele frequency'
      df['C_to_T_ratio'] = df['T'] / (df['C'] + df['T'] + 0.01)
      df['G_to_A_ratio'] = df['A'] / (df['G'] + df['A'] + 0.01) 
    elif (args.mode == "count"):
      y_label = 'Allele count'

    df_C = df[df['ref_nt'] == 'C']
    df_G = df[df['ref_nt'] == 'G']

    _, ax = plt.subplots()
    plot_kwargs = dict(edgecolor = 'black', linewidth = 0.5, s = 12, alpha = 0.75)
    
    if args.mode == "ratio":
      sns.scatterplot(data = df_C, ax = ax, x = 'start', y = 'C_to_T_ratio', label = 'T/(C + T) at ref C sites', **plot_kwargs, color = '#0072B2',)
      sns.scatterplot(data = df_G, ax = ax, x = 'start', y = 'G_to_A_ratio', label = 'A/(A + G) at ref G sites', **plot_kwargs, color = '#D55E00')
    elif args.mode == "count":
      sns.scatterplot(data = df_C, ax = ax, x = 'start', y = 'T', label = 'T count at ref C sites', **plot_kwargs, color = '#0072B2')
      sns.scatterplot(data = df_G, ax = ax, x = 'start', y = 'A', label = 'A count at ref G sites', **plot_kwargs, color = '#D55E00')
    
    # Add moving average line to ax
    window_size = 10
    df_C = df_C.copy() # avoid SettingWithCopyWarning
    df_G = df_G.copy()
    if args.mode == "ratio":
      df_C['C_to_T_ratio_rolling_avg'] = df_C['C_to_T_ratio'].rolling(window = window_size, min_periods = 1).mean()
      df_G['G_to_A_ratio_rolling_avg'] = df_G['G_to_A_ratio'].rolling(window = window_size, min_periods = 1).mean()
      sns.lineplot(data = df_C, ax = ax, x = 'start', 
                  y = 'C_to_T_ratio_rolling_avg',
                  label = 'T/(C + T) at ref C sites (rolling avg)',
                  color = '#0072B2', linewidth = 0.4, alpha = 0.6)
      sns.lineplot(data = df_G, ax = ax, x = 'start', 
                  y = 'G_to_A_ratio_rolling_avg',
                  label = 'A/(A + G) at ref G sites (rolling avg)',
                  color = '#D55E00', linewidth = 0.4, alpha = 0.6)
    elif args.mode == "count":
      df_C['T_rolling_avg'] = df_C['T'].rolling(window = window_size, min_periods = 1).mean()
      df_G['A_rolling_avg'] = df_G['A'].rolling(window = window_size, min_periods = 1).mean()
      sns.lineplot(data = df_C, ax = ax, x = 'start', 
                  y = 'T_rolling_avg', 
                  label = 'T count at ref C sites (rolling avg)',
                  color = '#0072B2', linewidth = 0.4, alpha = 0.6)
      sns.lineplot(data = df_G, ax = ax, x = 'start', 
                  y = 'A_rolling_avg', 
                  label = 'A count at ref G sites (rolling avg)',
                  color = '#D55E00', linewidth = 0.4, alpha = 0.6)
    
    # Add GC% line to ax2 if args.add_gc is True
    if (args.add_gc):
      window_size = 25
      ax2 = ax.twinx()
      
      df['is_GC'] = df['ref_nt'].isin(['G', 'C'])
      df['smoothed_GC_percent'] = df['is_GC'].rolling(window = window_size, min_periods = 1).mean()
    
      sns.lineplot(data = df, ax = ax2, x = 'start', y = 'smoothed_GC_percent', color = '#000000', linewidth = 0.3, alpha = 0.45)
      ax2.set_ylabel('GC%')
    
    ax.set_xlabel('Position')
    ax.set_ylabel(y_label)
    plt.savefig(args.output)
  elif args.type == 'nt_qual':
    # Calculate qual ratio for all levels
    if args.mode == "ratio":
      y_label = 'Qualilty frequency'
      df['0_10_ratio'] = df['qual_0_10'] / (df['qual_0_10'] + df['qual_10_20'] + df['qual_20_30'] + df['qual_30_40'] + df['qual_40_above'])
      df['10_20_ratio'] = df['qual_10_20'] / (df['qual_0_10'] + df['qual_10_20'] + df['qual_20_30'] + df['qual_30_40'] + df['qual_40_above'])
      df['20_30_ratio'] = df['qual_20_30'] / (df['qual_0_10'] + df['qual_10_20'] + df['qual_20_30'] + df['qual_30_40'] + df['qual_40_above'])
      df['30_40_ratio'] = df['qual_30_40'] / (df['qual_0_10'] + df['qual_10_20'] + df['qual_20_30'] + df['qual_30_40'] + df['qual_40_above'])
      df['40_above_ratio'] = df['qual_40_above'] / (df['qual_0_10'] + df['qual_10_20'] + df['qual_20_30'] + df['qual_30_40'] + df['qual_40_above'])
      y_cols = ['0_10_ratio', '10_20_ratio', '20_30_ratio', '30_40_ratio', '40_above_ratio']
    elif args.mode == 'count':
      y_label = 'Quality count'
      y_cols = ['qual_0_10', 'qual_10_20', 'qual_20_30', 'qual_30_40', 'qual_40_above']
    
    _, ax = plt.subplots()
    plot_kwargs = dict(edgecolor = 'black', linewidth = 0.5, s = 12, alpha = 0.75)
    n_bins = 5
    cmap = plt.cm.get_cmap('summer')
    colors = [cmap(i) for i in np.linspace(0, 1, n_bins)][::-1]
    labels = ['0-10', '10-20', '20-30', '30-40', '40+']
    
    bottom = np.zeros(len(df))
    for y_col, color, label in zip(y_cols, colors, labels):
      ax.bar(df['start'], df[y_col], bottom = bottom, color = color, label = label, width = 1.0)
      bottom += df[y_col].values

    ax.set_xlabel('Position')
    ax.set_ylabel(y_label)
    ax.legend(title='Quality Bin')
    
    if (args.add_gc):
      window_size = 25
      ax2 = ax.twinx()
      
      df['is_GC'] = df['ref_nt'].isin(['G', 'C'])
      df['smoothed_GC_percent'] = df['is_GC'].rolling(window = window_size, min_periods = 1).mean()
    
      sns.lineplot(data = df, ax = ax2, x = 'start', y = 'smoothed_GC_percent', color = '#000000', linewidth = 0.3, alpha = 0.45)
      ax2.set_ylabel('GC%')
      
    elif args.add_qual_avg:
      window_size = 25
      ax2 = ax.twinx()
      df['qual_avg_rolling_avg'] = df['qual_avg'].rolling(window = 25, min_periods = 1).mean()
      sns.lineplot(data = df, ax = ax2, x = 'start', y = 'qual_avg_rolling_avg', color = '#000000', linewidth = 0.3, alpha = 0.45)
      ax2.set_ylabel('Qual Avg (rolling avg)')
    
    plt.savefig(args.output)


import os
from .utils import ensure_tsv_index, extract_range_from_tsv
import seaborn as sns
import numpy as np  
import matplotlib.pyplot as plt
import dask.dataframe as dd
import pandas as pd

def merge_overlapping_intervals(df):
  if df.empty:
    return pd.DataFrame(columns=['start', 'end'])
  
  df = df.reset_index(drop = True)
  merged = []
  current_start, current_end = df.loc[0, 'start'], df.loc[0, 'end']
  for i in range(1, len(df)):
    row = df.loc[i]
    if row['start'] <= current_end:
      current_end = max(current_end, row['end'])
    else:
      merged.append((current_start, current_end))
      current_start, current_end = row['start'], row['end']
  merged.append((current_start, current_end))
  return pd.DataFrame(merged, columns = ['start', 'end'])

def merge_nearby_intervals(df, window_size):
  if df.empty:
    return pd.DataFrame(columns = ['start', 'end'])

  merged = []
  current_start, current_end = df.loc[0, 'start'], df.loc[0, 'end']
  
  for i in range(1, len(df)):
    row = df.loc[i]
    if row['start'] - current_end <= window_size:
      current_end = max(current_end, row['end'])
    else:
      merged.append((current_start, current_end))
      current_start, current_end = row['start'], row['end']

  merged.append((current_start, current_end))
  
  return pd.DataFrame(merged, columns = ['start', 'end'])

def group_func_nt_qual(x, low_qual_cutoff = 20, frac_cutoff = 0.3, window_size = 20, merge_nearby_peaks = True):
  '''
  Each chr is a group, perform sliding window approach within each group.
  '''
  
  chr = x['chr'].iloc[0]
  print("Processing chromosome: ", chr)
  # min, max = x['start'].min(), x['start'].max()
  if low_qual_cutoff == 10:
    low_group = x['qual_0_10']
  elif low_qual_cutoff == 20:
    low_group = x['qual_0_10'] + x['qual_10_20']
  elif low_qual_cutoff == 30:
    low_group = x['qual_0_10'] + x['qual_10_20'] + x['qual_20_30']
  elif low_qual_cutoff == 40:
    low_group = x['qual_0_10'] + x['qual_10_20'] + x['qual_20_30'] + x['qual_30_40']
    
  x['low_frac'] = (low_group) / (x['qual_0_10'] + x['qual_10_20'] + x['qual_20_30'] + x['qual_30_40'] + x['qual_40_above'])
  x = x.sort_values('start').reset_index(drop = True)
  x['low_frac_rolling_avg'] = x['low_frac'].rolling(window = window_size, min_periods = 1).mean() 
  x = x[x['low_frac_rolling_avg'] >= frac_cutoff]
  df = merge_overlapping_intervals(x)
  if merge_nearby_peaks:
    df = merge_nearby_intervals(df, window_size * 2)

  df['chr'] = chr
  df = df[['chr', 'start', 'end']]
  return df

def group_func_nt_count(x, conversion_cutoff = 0.1, window_size = 20, merge_nearby_peaks = True):
    '''
    Each chr is a group, perform sliding window approach within each group.
    '''
    
    chr = x['chr'].iloc[0]
    print("Processing chromosome: ", chr)
    
    x = x.copy()
    c_positions = x[x['ref_nt'] == 'C'].index.tolist()
    
    # For each non-C position, find the next C position and use its values
    for i in range(len(x)):
      if x.loc[i, 'ref_nt'] != 'C':
        next_c_pos = None
        for c_pos in c_positions:
          if c_pos > i:
            next_c_pos = c_pos
            break
    
        if next_c_pos is not None:
          x.loc[i, 'C'] = x.loc[next_c_pos, 'C']
          x.loc[i, 'T'] = x.loc[next_c_pos, 'T']
          
    # x = x[x['ref_nt'] == 'C']
    x['conversion_frac'] = x['T'] / (x['T'] + x['C'] + 0.01)
    x['conversion_frac_rolling_avg'] = x['conversion_frac'].rolling(window = window_size, min_periods = 1).mean() 
    x = x[x['conversion_frac_rolling_avg'] >= conversion_cutoff]
    
    df = merge_overlapping_intervals(x)
    if merge_nearby_peaks:
      df = merge_nearby_intervals(df, window_size * 2)
    df['chr'] = chr
    df = df[['chr', 'start', 'end']]
    return df

def run_tsv_to_peak(args):
  """
  Given tsv.gz from bam_to_tsv, call peaks using a sliding window approach.
  """
  
  # Check input
  if not os.path.exists(args.tsv):
    raise FileNotFoundError(f"The input TSV file {args.tsv} does not exist.")

  import dask
  dask.config.set(scheduler = 'processes', num_workers = args.ncpus) # parallel groupby may take large memory
  
  if args.type == 'nt_qual':
    print("Reading tsv.gz ...")
    cols = ['chr', 'start', 'end', 'ref_nt', 'qual_0_10', 'qual_10_20', 'qual_20_30', 'qual_30_40', 'qual_40_above', 'qual_avg']
    dtypes = {'chr': str, 'start': int, 'end': int, 'ref_nt': str, 'qual_0_10': int, 'qual_10_20': int, 'qual_20_30': int, 'qual_30_40': int, 'qual_40_above': int, 'qual_avg': float}
    
    if args.tsv.endswith('.gz'):
      df = dd.read_csv(args.tsv, compression = 'gzip', sep = '\t', names = cols, dtype = dtypes, header = None, blocksize = None, skiprows = 1)
    elif args.tsv.endswith('.tsv'):
      df = dd.read_csv(args.tsv, sep = '\t', names = cols, dtype = dtypes, header = None, blocksize = '500MB', skiprows = 1)
    
    # Group by chromosome and apply the rolling average window approach to call peaks
    df_res = df.groupby('chr').apply(group_func_nt_qual, low_qual_cutoff = args.low_qual_cutoff, frac_cutoff = args.frac_cutoff, window_size = args.window_size, merge_nearby_peaks = args.merge_nearby_peaks, meta = {'chr': str, 'start': int, 'end': int}).compute()
  elif args.type == 'nt_count':
    print("Reading tsv.gz ...")
    cols = ['chr', 'start', 'end', 'ref_nt', 'A', 'T', 'C', 'G', 'N']
    dtypes = {'chr': str, 'start': int, 'end': int, 'ref_nt': str, 'A': int, 'T': int, 'C': int, 'G': int, 'N': int}
    
    if args.tsv.endswith('.gz'):
      df = dd.read_csv(args.tsv, compression = 'gzip', sep = '\t', names = cols, dtype = dtypes, header = None, blocksize = None, skiprows = 1)
    elif args.tsv.endswith('.tsv'):
      df = dd.read_csv(args.tsv, sep = '\t', names = cols, dtype = dtypes, header = None, blocksize = '500MB', skiprows = 1)
    
    # Group by chromosome and apply the rolling average window approach to call peaks
    df_res = df.groupby('chr').apply(group_func_nt_count, conversion_cutoff = args.conversion_cutoff, window_size = args.window_size, merge_nearby_peaks = args.merge_nearby_peaks, meta = {'chr': str, 'start': int, 'end': int}).compute()

  # Drop peaks that are too short 
  df_res = df_res[df_res['end'] - df_res['start'] >= args.min_peak_length]
  df_res.to_csv(args.output, sep = '\t', index = False, compression = 'gzip', header = False)
  print("Peaks called and saved to ", args.output, sep = '')


import pysam
import os
import io
from statistics import mean
import pandas as pd
import numpy as np
from statistics import stdev
import re
from collections import Counter
import gzip

def validate_range(user_range):
    range_pattern = r"^(\w+):(\d+)-(\d+)$"
    match = re.match(range_pattern, user_range)
    if match == None:
      raise ValueError(f"--range format is incorrect. Expected format: chrname:start-end")
    
    _, start, end = match.groups()
    start, end = int(start), int(end)
    if start >= end:
      raise ValueError(f"--range: start position must be less than end position.")

def ensure_bam_index(bam_file):
    bam_index = bam_file + '.bai'
    if not os.path.exists(bam_index):
        print(f"Index not found for {bam_file}. Creating index ...")
        pysam.index(bam_file)
        
def ensure_tsv_index(tsv_file):
    tsv_index = tsv_file + '.tbi'
    if not os.path.exists(tsv_index):
        print(f"Index not found for {tsv_file}. Creating index ...")
        pysam.tabix_index(tsv_file, preset = "bed", force = True)
        
def parse_filename(path):
    filename = os.path.basename(path)
    match = re.match(r"([a-zA-Z0-9]+)_([\d]+)_([\d]+)", filename)
    if match:
        chr = match.group(1)
        start = int(match.group(2))
        return chr, start
    return None, None

def chunk_bam(ref_name, ref_length, num_chunks):
  chunk_size = (ref_length + num_chunks - 1) // num_chunks 
  chunks = [(ref_name, start, min(start + chunk_size, ref_length)) 
            for start in range(0, ref_length, chunk_size)]  
  return chunks

def extract_range_from_tsv(file_path, region):
  with gzip.open(file_path, 'rt') as f:
    header = f.readline().strip() 
    if header.startswith('#'):
      header = header[1:].split('\t')
    else:
      n_col = len(header.split('\t'))
      header = [f'col{i + 1}' for i in range(n_col)]

  with pysam.TabixFile(file_path) as tabix_file:
    print(file_path)
    lines = list(tabix_file.fetch(region = region))
    if not lines:
      raise ValueError('Warnning: no data found in the specified region.')
      
    data = '\n'.join(lines)

    df = pd.read_csv(io.StringIO(data), sep = '\t', header = None)
    df.columns = header 
    
    return df

def process_bam_chunk_nt_qual(ref_name, start_pos, end_pos, bam_path, ref_path, temp_dir):
  print(f"Processing chunk: {ref_name}_{start_pos}_{end_pos}")

  results = []
  with pysam.AlignmentFile(bam_path, "rb") as bam, pysam.FastaFile(ref_path) as ref:
    temp_file = os.path.join(temp_dir, '{}_{}_{}.tsv'.format(ref_name, start_pos, end_pos))
  
    for pileup_column in bam.pileup(ref_name, start_pos, end_pos, min_base_quality = 0, max_depth = 2_147_483_647):
      pos = pileup_column.pos
      if pos < start_pos or pos >= end_pos:
        continue 
      
      ref_base = ref.fetch(ref_name, pos, pos + 1).upper()
      ref_position = pos # 0-based
      
      quals = []
      for pileup_read in pileup_column.pileups:
        read_pos = pileup_read.query_position
        read_qual = pileup_read.alignment.query_qualities[read_pos] if read_pos is not None else 0 
        quals.append(read_qual)

      bins = [0, 10, 20, 30, 40, np.inf]
      bin_labels = ['qual_0_10', 'qual_10_20', 'qual_20_30', 'qual_30_40', 'qual_40_above']
      bin = pd.cut(quals, bins = bins, labels = bin_labels, right = False)
      
      results.append(
          {
            'chr': ref_name,
            'start': ref_position,
            'end': ref_position + 1,
            'ref_base': ref_base,
            bin_labels[0]: bin.value_counts()[bin_labels[0]],
            bin_labels[1]: bin.value_counts()[bin_labels[1]],
            bin_labels[2]: bin.value_counts()[bin_labels[2]],
            bin_labels[3]: bin.value_counts()[bin_labels[3]],
            bin_labels[4]: bin.value_counts()[bin_labels[4]],
            'qual_avg': np.mean(quals)
          }
      )
      
      if len(results) >= 1000000:
        df = pd.DataFrame(results)
        if not pd.io.common.file_exists(temp_file):
            df.to_csv(temp_file, sep = '\t', index = False, mode = 'w', header = False)
        else:
            df.to_csv(temp_file, sep = '\t', index = False, mode = 'a', header = False)
        results = []

  df = pd.DataFrame(results)
  if not pd.io.common.file_exists(temp_file):
    df.to_csv(temp_file, sep = '\t', index= False, mode = 'w', header = False)
  else:
    df.to_csv(temp_file, sep = '\t', index = False, mode = 'a', header = False)
  
  return (temp_file)

def process_bam_chunk_nt_count(ref_name, start_pos, end_pos, bam_path, ref_path, temp_dir):
  print(f"Processing chunk: {ref_name}_{start_pos}_{end_pos}")

  results = []
  with pysam.AlignmentFile(bam_path, "rb") as bam, pysam.FastaFile(ref_path) as ref:
    temp_file = os.path.join(temp_dir, '{}_{}_{}.tsv'.format(ref_name, start_pos, end_pos))
  
    for pileup_column in bam.pileup(ref_name, start_pos, end_pos, min_base_quality = 0, max_depth = 2_147_483_647):
      pos = pileup_column.pos
      if pos < start_pos or pos >= end_pos:
        continue 
      
      ref_base = ref.fetch(ref_name, pos, pos + 1).upper()
      ref_position = pos # 0-based
      
      nt = []
      for pileup_read in pileup_column.pileups:
        read_pos = pileup_read.query_position
        read_nt = pileup_read.alignment.query_sequence[read_pos].upper() if read_pos is not None else 'N'
        nt.append(read_nt)

      bins = ['A', 'T', 'C', 'G', 'N']
      counts = Counter(nt)
      bin = {nt: counts.get(nt, 0) for nt in bins}
            
      results.append(
          {
            'chr': ref_name,
            'start': ref_position,
            'end': ref_position + 1,
            'ref_base': ref_base,
            bins[0]: bin[bins[0]],
            bins[1]: bin[bins[1]],
            bins[2]: bin[bins[2]],
            bins[3]: bin[bins[3]],
            bins[4]: bin[bins[4]]
          }
      )
      
      if len(results) >= 1000000:
        df = pd.DataFrame(results)
        if not pd.io.common.file_exists(temp_file):
            df.to_csv(temp_file, sep = '\t', index = False, mode = 'w', header = False)
        else:
            df.to_csv(temp_file, sep = '\t', index = False, mode = 'a', header = False)
        results = []

  df = pd.DataFrame(results)
  if not pd.io.common.file_exists(temp_file):
    df.to_csv(temp_file, sep = '\t', index= False, mode = 'w', header = False)
  else:
    df.to_csv(temp_file, sep = '\t', index = False, mode = 'a', header = False)
  
  return (temp_file)



import os
from .utils import ensure_tsv_index, extract_range_from_tsv
import seaborn as sns
import numpy as np  
import matplotlib.pyplot as plt
import gzip
import pandas as pd
def run_tsv_to_bed(args):
  """
  Given tsv.gz from bam_to_tsv, parse out bed-like format for MACS2 peak calling.
  """
  
  # Check input
  if not os.path.exists(args.tsv):
    raise FileNotFoundError(f"The input TSV file {args.tsv} does not exist.")
  ensure_tsv_index(args.tsv)

  if args.type == 'nt_qual':
    with gzip.open(args.output, 'wt') as out:
      with gzip.open(args.tsv, 'rt') as f:
        header = f.readline()[1:].strip().split('\t')
        chunk_iter = pd.read_csv(f, sep = '\t', header = None, names = header, chunksize = 10000)
        for chunk in chunk_iter:
          for _, row in chunk.iterrows():
            qual_avg = row['qual_avg']
            qual_avg = 60 - qual_avg # 60 is the max quality score theoretically
            qual_avg = qual_avg * args.scale
            qual_avg = int(qual_avg)
            if qual_avg < 0:
              qual_avg = 0
            bed_line = f"{row['chr']}\t{row['start']}\t{row['end']}\n"
            for _ in range(qual_avg):
              out.write(bed_line)

  elif args.type == 'nt_count':
    with gzip.open(args.output, 'wt') as out:
      with gzip.open(args.tsv, 'rt') as f:
        header = f.readline()[1:].strip().split('\t')
        chunk_iter = pd.read_csv(f, sep = '\t', header = None, names = header, chunksize = 10000)
        for chunk in chunk_iter:
          for _, row in chunk.iterrows():
            if row['ref_nt'] == 'C':
              row['C_to_T_ratio'] = row['T'] / (row['C'] + row['T'] + 0.01) * 100 * args.scale 
              row['C_to_T_ratio'] = int(row['C_to_T_ratio'])
              bed_line = f"{row['chr']}\t{row['start']}\t{row['end']}\n"
              for _ in range(row['C_to_T_ratio']):
                out.write(bed_line)